@@ -2,13 +2,13 @@
 // Program to pull the information out of various types of EXIF digital 
 // camera files and show it in a reasonably consistent way
 //
-// Version 2.4-2
+// Version 2.6
 //
 //
-// Compiling under Windows:  Use microsoft's compiler.  from command line:
-// cl -Ox jhead.c exif.c myglob.c
+// Compiling under Windows:  
+//   Make sure you have microsoft's compiler on the path, then run make.bat
 //
-// Dec 1999 - Jun 2005
+// Dec 1999 - April 2006
 //
 // by Matthias Wandel   www.sentex.net/~mwandel
 //--------------------------------------------------------------------------
@@ -22,7 +22,7 @@
 #include <errno.h>
 #include <ctype.h>
 
-#define JHEAD_VERSION "2.44"
+#define JHEAD_VERSION "2.6"
 
 // This #define turns on features that are too very specific to 
 // how I organize my photos.  Best to ignore everything inside #ifdef MATTHIAS
@@ -53,24 +53,32 @@ static const char * progname;   // program name for error messages
 // Command line options flags
 static int TrimExif = FALSE;        // Cut off exif beyond interesting data.
 static int RenameToDate = FALSE;
+static int RenameAssociatedFiles = FALSE;
 static char * strftime_args = NULL; // Format for new file name.
 static int Exif2FileTime  = FALSE;
 static int DoModify     = FALSE;
 static int DoReadAction = FALSE;
        int ShowTags     = FALSE;    // Do not show raw by default.
+       int DumpExifMap  = FALSE;
 static int ShowConcise  = FALSE;
 static char * ApplyCommand = NULL;  // Apply this command to all images.
 static char * FilterModel = NULL;
 static int    ExifOnly    = FALSE;
+static int    PortraitOnly = 0;
 static time_t ExifTimeAdjust = 0;   // Timezone adjust
 static time_t ExifTimeSet = 0;      // Set exif time to a value.
 
 static int DeleteComments = FALSE;
 static int DeleteExif = FALSE;
 static int DeleteUnknown = FALSE;
-static char * ThumbnailName = NULL; // If not NULL, use this string to make up
+static char * ThumbSaveName = NULL; // If not NULL, use this string to make up
                                     // the filename to store the thumbnail to.
 
+static char * ThumbInsertName = NULL; // If not NULL, use this string to make up
+                                    // the filename to retrieve the thumbnail from.
+
+static int RegenThumbnail = 0;
+
 static char * ExifXferScrFile = NULL;// Extract Exif header from this file, and
                                     // put it into the Jpegs processed.
 
@@ -131,7 +139,7 @@ static int FileEditComment(char * TempFileName, char * Comment, int CommentSize)
 {
     FILE * file;
     int a;
-    char QuotedPath[300];
+    char QuotedPath[PATH_MAX];
 
     file = fopen(TempFileName, "w");
     if (file == NULL){
@@ -184,10 +192,9 @@ static int FileEditComment(char * TempFileName, char * Comment, int CommentSize)
 // Modify one of the lines in the comment field.
 // This very specific to the photo album program stuff.
 //--------------------------------------------------------------------------
-static char KnownTags[][10] = {"date", "orig_path", "desc","scan_date","author",
-                               "mwnum", "crop", "rotate", "subpic", 
-                               "related", "infopic", "keyword","videograb",
-                               "show_raw","panorama","uid",""};
+static char KnownTags[][10] = {"date", "desc","scan_date","author",
+                               "keyword","videograb",
+                               "show_raw","panorama","titlepix",""};
 
 static int ModifyDescriptComment(char * OutComment, char * SrcComment)
 {
@@ -299,7 +306,7 @@ static int AutoResizeCmdStuff(void)
 //--------------------------------------------------------------------------
 // Apply the specified command to the JPEG file.
 //--------------------------------------------------------------------------
-static void DoCommand(const char * FileName)
+static void DoCommand(const char * FileName, int ShowIt)
 {
     int a,e;
     char ExecString[400];
@@ -335,7 +342,7 @@ static void DoCommand(const char * FileName)
         if (ApplyCommand[a] == 0) break;
     }
 
-    printf("Cmd:%s\n",ExecString);
+    if (ShowIt) printf("Cmd:%s\n",ExecString);
 
     errno = 0;
     a = system(ExecString);
@@ -368,6 +375,8 @@ static int CheckFileSkip(void)
 
     if (FilterModel){
         // Filtering processing by camera model.
+        // This feature is useful when pictures from multiple cameras are colated, 
+        // the its found that one of the cameras has the time set incorrectly.
         if (strstr(ImageInfo.CameraModel, FilterModel) == NULL){
             // Skip.
             return TRUE;
@@ -381,6 +390,14 @@ static int CheckFileSkip(void)
         }
     }
 
+    if (PortraitOnly == 1){
+        if (ImageInfo.Width > ImageInfo.Height) return TRUE;
+    }
+
+    if (PortraitOnly == -1){
+        if (ImageInfo.Width < ImageInfo.Height) return TRUE;
+    }
+
     return FALSE;
 }
 
@@ -406,11 +423,65 @@ static void RelativeName(char * OutFileName, const char * NamePattern, const cha
 }
 
 
+#ifdef _WIN32
+//--------------------------------------------------------------------------
+// Rename associated files
+//--------------------------------------------------------------------------
+void RenameAssociated(const char * FileName, char * NewBaseName)
+{
+    int a;
+    int PathLen;
+    int ExtPos;
+    char FilePattern[_MAX_PATH];
+    char NewName[_MAX_PATH];
+    struct _finddata_t finddata;
+    long find_handle;
+
+    for(ExtPos = strlen(FileName);FileName[ExtPos-1] != '.';){
+        if (--ExtPos == 0) return; // No extension!
+    }
+
+    memcpy(FilePattern, FileName, ExtPos);
+    FilePattern[ExtPos] = '*';
+    FilePattern[ExtPos+1] = '\0';
+
+    for(PathLen = strlen(FileName);FileName[PathLen-1] != '\\';){
+        if (--PathLen == 0) break;
+    }
+
+    find_handle = _findfirst(FilePattern, &finddata);
+
+    for (;;){
+        if (find_handle == -1) break;
+
+        // Eliminate the obvious patterns.
+        if (!memcmp(finddata.name, ".",2)) goto next_file;
+        if (!memcmp(finddata.name, "..",3)) goto next_file;
+        if (finddata.attrib & _A_SUBDIR) goto next_file;
+
+        strcpy(FilePattern+PathLen, finddata.name); // full name with path
+
+        strcpy(NewName, NewBaseName);
+        for(a = strlen(finddata.name);finddata.name[a] != '.';){
+            if (--a == 0) goto next_file;
+        }
+        strcat(NewName, finddata.name+a); // add extension to new name
+
+        if (rename(FilePattern, NewName) == 0){
+            printf("%s --> %s\n",FilePattern, NewName);
+        }
+
+        next_file:
+        if (_findnext(find_handle, &finddata) != 0) break;
+    }
+    _findclose(find_handle);
+}
+#endif
 
 //--------------------------------------------------------------------------
 // Handle renaming of files by date.
 //--------------------------------------------------------------------------
-void DoFileRenaming(const char * FileName)
+static void DoFileRenaming(const char * FileName)
 {
     int NumAlpha = 0;
     int NumDigit = 0;
@@ -535,6 +606,12 @@ void DoFileRenaming(const char * FileName)
             // This name does not pre-exist.
             if (rename(FileName, NewName) == 0){
                 printf("%s --> %s\n",FileName, NewName);
+#ifdef _WIN32
+                if (RenameAssociatedFiles){
+                    sprintf(NewName, "%s%s", NewBaseName, NameExtra);
+                    RenameAssociated(FileName, NewName);
+                }
+#endif
             }else{
                 printf("Error: Couldn't rename '%s' to '%s'\n",FileName, NewName);
             }
@@ -548,6 +625,79 @@ void DoFileRenaming(const char * FileName)
     }
 }
 
+//--------------------------------------------------------------------------
+// Rotate the image and its thumbnail
+//--------------------------------------------------------------------------
+static int DoAutoRotate(const char * FileName)
+{
+    if (ImageInfo.Orientation >= 2 && ImageInfo.Orientation <= 8){
+        const char * Argument;
+        Argument = ClearOrientation();
+
+        if (!ZeroRotateTagOnly){
+            char RotateCommand[PATH_MAX*2+50];
+            if (Argument == NULL){
+                ErrFatal("Orientation screwup");
+            }
+
+            sprintf(RotateCommand, "jpegtran -%s -outfile &o &i", Argument);
+            ApplyCommand = RotateCommand;
+            DoCommand(FileName, FALSE);
+            ApplyCommand = NULL;
+
+            // Now rotate the thumbnail, if there is one.
+            if (ImageInfo.ThumbnailOffset && 
+                ImageInfo.ThumbnailSize && 
+                ImageInfo.ThumbnailAtEnd){
+                // Must have a thumbnail that exists and is modifieable.
+
+                char ThumbTempName_in[PATH_MAX+4];
+                char ThumbTempName_out[PATH_MAX+4];
+
+                strcpy(ThumbTempName_in, FileName);
+                strcat(ThumbTempName_in, ".thi");
+                strcpy(ThumbTempName_out, FileName);
+                strcat(ThumbTempName_out, ".tho");
+                SaveThumbnail(ThumbTempName_in);
+                sprintf(RotateCommand,"jpegtran -%s -outfile \"%s\" \"%s\"",
+                    Argument, ThumbTempName_out, ThumbTempName_in);
+
+                if (system(RotateCommand) == 0){
+                    // Put the thumbnail back in the header
+                    ReplaceThumbnail(ThumbTempName_out);
+                }
+
+                unlink(ThumbTempName_in);
+                unlink(ThumbTempName_out);
+            }
+        }
+        return TRUE;
+    }
+    return FALSE;
+}
+
+//--------------------------------------------------------------------------
+// Regenerate the thumbnail using mogrify
+//--------------------------------------------------------------------------
+static int RegenerateThumbnail(const char * FileName)
+{
+    char ThumbnailGenCommand[PATH_MAX*2+50];
+    if (ImageInfo.ThumbnailOffset == 0 || ImageInfo.ThumbnailAtEnd == FALSE){
+        // There is no thumbnail, or the thumbnail is not at the end.
+        return FALSE;
+    }
+
+    sprintf(ThumbnailGenCommand, "mogrify -thumbnail %dx%d \"%s\"", 
+        RegenThumbnail, RegenThumbnail, FileName);
+
+    if (system(ThumbnailGenCommand) == 0){
+        // Put the thumbnail back in the header
+        return ReplaceThumbnail(FileName);
+    }else{
+        ErrFatal("Unable to run 'mogrify' command");
+        return FALSE;
+    }
+}
 
 //--------------------------------------------------------------------------
 // Do selected operations to one file at a time.
@@ -613,25 +763,18 @@ void ProcessFile(const char * FileName)
         DiscardAllButExif();
 
         if (AutoRotate){
-            if (ImageInfo.Orientation >= 2 && ImageInfo.Orientation <= 8){
-                const char * Argument;
-                Argument = ClearOrientation();
-
-                if (!ZeroRotateTagOnly){
-                    char RotateCommand[50];
-                    if (Argument == NULL){
-                        ErrFatal("Orientation screwup");
-                    }
-                    sprintf(RotateCommand, "jpegtran -%s -outfile &o &i", Argument);
-                    ApplyCommand = RotateCommand;
-                    DoCommand(FileName);
-                    ApplyCommand = NULL;
-                }
-
+            if (DoAutoRotate(FileName)){
                 Modified = TRUE;
             }
         }else{
-            DoCommand(FileName);
+            struct stat dummy;
+            DoCommand(FileName, TRUE);
+
+            if (stat(FileName, &dummy)){
+                // The file is not there anymore. Perhaps the command
+                // was a delete or a move.  So we are all done.
+                return;
+            }
             Modified = TRUE;
         }
         ReadMode = READ_IMAGE;   // Don't re-read exif section again on next read.
@@ -671,38 +814,36 @@ void ProcessFile(const char * FileName)
         }
     }
 
-    if (ThumbnailName){
-        if (ImageInfo.ThumbnailPointer){
-            FILE * ThumbnailFile;
-            char OutFileName[PATH_MAX+1];
+    if (ThumbSaveName){
+        char OutFileName[PATH_MAX+1];
+        // Make a relative name.
+        RelativeName(OutFileName, ThumbSaveName, FileName);
 
-            // Make a relative name.
-            RelativeName(OutFileName, ThumbnailName, FileName);
-
-#ifndef _WIN32
-            if (strcmp(ThumbnailName, "-") == 0){
-                // A filename of '-' indicates thumbnail goes to stdout.
-                // This doesn't make much sense under Windows, so this feature is unix only.
-                ThumbnailFile = stdout;
-            }else
-#endif
-            {
-                ThumbnailFile = fopen(OutFileName,"wb");
+        if (SaveThumbnail(OutFileName)){
+            if (OutFileName != "-"){
+                printf("Created: '%s'\n", OutFileName);
             }
+        }
+    }
 
-            if (ThumbnailFile){
-                fwrite(ImageInfo.ThumbnailPointer, ImageInfo.ThumbnailSize ,1, ThumbnailFile);
-                fclose(ThumbnailFile);
-                if (ThumbnailFile != stdout){
-                    printf("Created: '%s'\n", OutFileName);
-                }else{
-                    // No point in printing to stdout when that is where the thumbnail goes!
-                }
-            }else{
-                ErrFatal("Could not write thumbnail file");
-            }
-        }else{
-            printf("Image '%s' contains no thumbnail\n",FileName);
+    if (RegenThumbnail){
+        if (RegenerateThumbnail(FileName)){
+            Modified = TRUE;
+        }
+    }
+
+    if (ThumbInsertName){
+        char ThumbFileName[PATH_MAX+1];
+        // Make a relative name.
+        RelativeName(ThumbFileName, ThumbInsertName, FileName);
+
+        if (ReplaceThumbnail(ThumbFileName)){
+            Modified = TRUE;
+        }
+    }else if (TrimExif){
+        // Deleting thumbnail is just replacing it with a null thumbnail.
+        if (ReplaceThumbnail(NULL)){
+            Modified = TRUE;
         }
     }
 
@@ -826,6 +967,7 @@ void ProcessFile(const char * FileName)
             time_t UnixTime;
             char TempBuf[50];
             int a;
+            Section_t * ExifSection;
 
             if (ExifTimeSet){
                 // A time to set was specified.
@@ -847,8 +989,12 @@ void ProcessFile(const char * FileName)
                 tm.tm_year+1900, tm.tm_mon+1, tm.tm_mday,
                 tm.tm_hour, tm.tm_min, tm.tm_sec);
 
+            ExifSection = FindSection(M_EXIF);
+
             for (a = 0; a < ImageInfo.numDateTimeTags; a++) {
-                memcpy(ImageInfo.DateTimePointers[a], TempBuf, 19);
+                uchar * Pointer;
+                Pointer = ExifSection->Data+ImageInfo.DateTimeOffsets[a]+8;
+                memcpy(Pointer, TempBuf, 19);
             }
 
             Modified = TRUE;
@@ -857,10 +1003,6 @@ void ProcessFile(const char * FileName)
         }
     }
 
-    if (TrimExif){
-        if (TrimExifFunc()) Modified = TRUE;
-    }
-    
     if (DeleteComments){
         if (RemoveSectionType(M_COM)) Modified = TRUE;
     }
@@ -874,6 +1016,8 @@ void ProcessFile(const char * FileName)
 
     if (Modified){
         char BackupName[400];
+        struct stat buf;
+
         printf("Modified: %s\n",FileName);
 
         strcpy(BackupName, FileName);
@@ -888,6 +1032,18 @@ void ProcessFile(const char * FileName)
         // Write the new file.
         WriteJpegFile(FileName);
 
+        // Copy the access rights from original file
+        if (stat(BackupName, &buf) == 0){
+            // set Unix access rights and time to new file
+            struct utimbuf mtime;
+            chmod(FileName, buf.st_mode);
+
+            mtime.actime = buf.st_mtime;
+            mtime.modtime = buf.st_mtime;
+            
+            utime(FileName, &mtime);
+        }
+
         // Now that we are done, remove original file.
         unlink(BackupName);
     }
@@ -938,14 +1094,19 @@ badtime:
 //--------------------------------------------------------------------------
 static void Usage (void)
 {
-    printf("Program for extracting Digicam setting information from Exif JPEG headers\n"
-           "used by most Digital Cameras.  v"JHEAD_VERSION" Matthias Wandel, Jun 09 2005.\n"
+    printf("Jhead is a program for manipulating settings and thumnails in Exif jpeg headers\n"
+           "used by most Digital Cameras.  v"JHEAD_VERSION" Matthias Wandel, April 29 2006.\n"
            "http://www.sentex.net/~mwandel/jhead\n"
            "\n");
 
     printf("Usage: %s [options] files\n", progname);
     printf("Where:\n"
+           " files       path/filenames with or without wildcards\n"
+
            "[options] are:\n"
+           "\nGENERAL METADATA:\n"
+           "  -te <name> Transfer exif header from another image file <name>\n"
+           "             Uses same name mangling as '-st' option\n"
            "  -dc        Delete comment field (as left by progs like Photoshop & Compupic)\n"
            "  -de        Strip Exif section (smaller JPEG file, but lose digicam info)\n"
            "  -du        Delete non image sections except for Exif and comment sections\n"
@@ -957,36 +1118,8 @@ static void Usage (void)
            "  -ci <name> Insert comment section from a file.  -cs and -ci use same naming\n"
            "             scheme as used by the -st option\n"
            "  -cl string Insert literal comment string\n"
-           "  -autorot   Invoke jpegtran to rotate images according to Exif orientation tag\n"
-           "             Note: Windows users must get jpegtran for this to work\n"
-           "  -norot     Zero out the rotation tag.  This to avoid some browsers from\n" 
-           "             rotating the image again after you rotated it but neglected to\n"
-           "             clear the rotation tag\n"
-#ifdef MATTHIAS
-           "  -cr        Remove comment tag (my way)\n"
-           "  -ca        Add comment tag (my way)\n"
-           "  -ar        Auto resize to fit in 800x800, but never less than half\n"
-#endif //MATTHIAS
-           "  -st <name> Save Exif thumbnail, if there is one, in file <name>\n"
-           "             If output file name contains the substring \"&i\" then the\n"
-           "             image file name is substitute for the &i.  Note that quotes around\n"
-           "             the argument are required for the '&' to be passed to the program.\n"
-#ifndef _WIN32
-           "             An output name of '-' causes thumbnail to be written to stdout\n"
-#endif
-           "  -te <name> Transfer exif header from another image file <name>\n"
-           "             Uses same name mangling as '-st' option\n"
-           "  -dt        Remove exif integral thumbnails.   Typically trims 10k\n"
-           "  -h         help (this text)\n"
-           "  -v         even more verbose output\n"
-           "  -se        Supress error messages relating to corrupt exif header structure\n"
-           "  -c         concise output\n"
-           "  -nofinfo   Don't show file info (name/size/date)\n"
-           "  -model model\n"
-           "             Only process files from digicam containing model substring in\n"
-           "             camera model description\n"
-           "  -exonly    Skip all files that don't have an exif header (skip all jpegs that\n"
-           "             were not created by digicam)\n"
+
+           "\nDATE / TIME MANIPULATION:\n"
            "  -ft        Set file modification time to Exif time.\n"
            "  -n[format-string]\n"
            "             Rename files according to date.  Uses exif date if present, file\n"
@@ -1005,6 +1138,9 @@ static void Usage (void)
            "             the end of the name to make it unique.\n"
            "  -nf[format-string]\n"
            "             Same as -n, but rename regardless of original name\n"
+           "  -a         (Windows only) Rename files with same name but different extension\n"
+           "             Use together with -n to rename .AVI files from exif in .THM files\n"
+           "             for example\n"
            "  -ta<+|->h[:mm[:ss]]\n"
            "             Adjust time by h:mm backwards or forwards.  Useful when having\n"
            "             taken pictures with the wrong time set on the camera, such as when\n"
@@ -1022,6 +1158,44 @@ static void Usage (void)
            "             and time in the format yyyy:mmm:dd/hh:mm:ss\n"
            "  -ts<time>  Set the Exif internal time to <time>.  <time> is in the format\n"
            "             yyyy:mm:dd-hh:mm:ss\n"
+
+           "\nTHUMBNAIL MANIPULATION:\n"
+           "  -dt        Remove exif integral thumbnails.   Typically trims 10k\n"
+           "  -st <name> Save Exif thumbnail, if there is one, in file <name>\n"
+           "             If output file name contains the substring \"&i\" then the\n"
+           "             image file name is substitute for the &i.  Note that quotes around\n"
+           "             the argument are required for the '&' to be passed to the program.\n"
+#ifndef _WIN32
+           "             An output name of '-' causes thumbnail to be written to stdout\n"
+#endif
+           "  -rt <name> Replace Exif thumbnail.  Can only be done with headers that\n"
+           "             already contain a thumbnail.\n"
+           "  -rgt[size] Regnerate exif thumbnail.  Only works if image already\n"
+           "             contains a thumbail.  size specifies maximum height or width of\n"
+           "             thumbnail.  Relies on 'mogrify' programs to be on path\n"
+
+           "\nROTATION TAG MANIPULATION:\n"
+           "  -autorot   Invoke jpegtran to rotate images according to Exif orientation tag\n"
+           "             Note: Windows users must get jpegtran for this to work\n"
+           "  -norot     Zero out the rotation tag.  This to avoid some browsers from\n" 
+           "             rotating the image again after you rotated it but neglected to\n"
+           "             clear the rotation tag\n"
+
+           "\nOUTPUT VERBOSITY CONTROL:\n"
+           "  -h         help (this text)\n"
+           "  -v         even more verbose output\n"
+           "  -V         Show jhead version\n"
+           "  -exifmap   Dump header bytes, annotate.  Pipe thru sort for better viewing\n"
+           "  -se        Supress error messages relating to corrupt exif header structure\n"
+           "  -c         concise output\n"
+           "  -nofinfo   Don't show file info (name/size/date)\n"
+
+           "\nFILE MATCHING AND SELECTION:\n"
+           "  -model model\n"
+           "             Only process files from digicam containing model substring in\n"
+           "             camera model description\n"
+           "  -exonly    Skip all files that don't have an exif header (skip all jpegs that\n"
+           "             were not created by digicam)\n"
            "  -cmd command\n"
            "             Apply 'command' to every file, then re-insert exif and command\n"
            "             sections into the image. &i will be substituted for the input file\n"
@@ -1037,6 +1211,8 @@ static void Usage (void)
            "                jhead -cmd \"jpegtran -progressive &i &o\" *.jpg\n"
            "             to convert jpegs to progressive jpegs (Unix jpegtran syntax\n"
            "             differs slightly)\n"
+           "  -orp       Only operate on 'portrait' aspect ratio images\n"
+           "  -orl       Only operate on 'landscape' aspect ratio images\n"
 #ifdef _WIN32
            "  -r         No longer supported.  Use the ** wildcard to recurse directories\n"
            "             with instead.\n"
@@ -1044,7 +1220,16 @@ static void Usage (void)
            "                 jhead **/*.jpg\n"
            "                 jhead \"c:\\my photos\\**\\*.jpg\"\n"
 #endif
-           " files       path/filenames with or without wildcards\n"
+
+
+#ifdef MATTHIAS
+           "\n"
+           "  -cr        Remove comment tag (my way)\n"
+           "  -ca        Add comment tag (my way)\n"
+           "  -ar        Auto resize to fit in 1024x1024, but never less than half\n"
+#endif //MATTHIAS
+
+
            );
 
     exit(EXIT_FAILURE);
@@ -1096,20 +1281,74 @@ int main (int argc, char **argv)
     for (argn=1;argn<argc;argn++){
         arg = argv[argn];
         if (arg[0] != '-') break; // Filenames from here on.
-        if (!strcmp(arg,"-v")){
+
+    // General metadata options:
+        if (!strcmp(arg,"-te")){
+            ExifXferScrFile = argv[++argn];
+            DoModify = TRUE;
+        }else if (!strcmp(arg,"-dc")){
+            DeleteComments = TRUE;
+            DoModify = TRUE;
+        }else if (!strcmp(arg,"-de")){
+            DeleteExif = TRUE;
+            DoModify = TRUE;
+        }else if (!strcmp(arg, "-du")){
+            DeleteUnknown = TRUE;
+            DoModify = TRUE;
+        }else if (!strcmp(arg, "-purejpg")){
+            DeleteExif = TRUE;
+            DeleteComments = TRUE;
+            DeleteUnknown = TRUE;
+            DoModify = TRUE;
+        }else if (!strcmp(arg,"-ce")){
+            EditComment = TRUE;
+            DoModify = TRUE;
+        }else if (!strcmp(arg,"-cs")){
+            CommentSavefileName = argv[++argn];
+            DoModify = TRUE;
+        }else if (!strcmp(arg,"-ci")){
+            CommentInsertfileName = argv[++argn];
+            DoModify = TRUE;
+        }else if (!strcmp(arg,"-cl")){
+            CommentInsertLiteral = argv[++argn];
+            DoModify = TRUE;
+
+    // Output verbosity control
+        }else if (!strcmp(arg,"-h")){
+            Usage();
+        }else if (!strcmp(arg,"-v")){
             ShowTags = TRUE;
         }else if (!strcmp(arg,"-V")){
             printf("Jhead version: "JHEAD_VERSION"   Compiled: "__DATE__"\n");
             exit(0);
+        }else if (!strcmp(arg,"-exifmap")){
+            DumpExifMap = TRUE;
+        }else if (!strcmp(arg,"-se")){
+            SupressNonFatalErrors = TRUE;
+        }else if (!strcmp(arg,"-c")){
+            ShowConcise = TRUE;
+        }else if (!strcmp(arg,"-nofinfo")){
+            ShowFileInfo = 0;
+
+    // Thumbnail manipulation options
         }else if (!strcmp(arg,"-dt")){
             TrimExif = TRUE;
             DoModify = TRUE;
         }else if (!strcmp(arg,"-st")){
-            ThumbnailName = argv[++argn];
+            ThumbSaveName = argv[++argn];
             DoReadAction = TRUE;
-        }else if (!strcmp(arg,"-te")){
-            ExifXferScrFile = argv[++argn];
+        }else if (!strcmp(arg,"-rt")){
+            ThumbInsertName = argv[++argn];
             DoModify = TRUE;
+        }else if (!memcmp(arg,"-rgt", 4)){
+            RegenThumbnail = 160;
+            sscanf(arg+4, "%d", &RegenThumbnail);
+            if (RegenThumbnail > 320){
+                ErrFatal("Specified thumbnail geometry too big!");
+            }
+            DoModify = TRUE;
+
+    // Rotation tag manipulation
         }else if (!strcmp(arg,"-autorot")){
             AutoRotate = 1;
             DoModify = TRUE;
@@ -1117,8 +1356,8 @@ int main (int argc, char **argv)
             AutoRotate = 1;
             ZeroRotateTagOnly = 1;
             DoModify = TRUE;
-        }else if (!strcmp(arg,"-nofinfo")){
-            ShowFileInfo = 0;
+
+    // Date/Time manipulation options
         }else if (!memcmp(arg,"-n",2)){
             RenameToDate = 1;
             DoReadAction = TRUE; // Rename doesn't modify file, so count as read action.
@@ -1132,6 +1371,11 @@ int main (int argc, char **argv)
                 strftime_args = arg;
                 //printf("strftime_args = %s\n",arg);
             }
+        }else if (!strcmp(arg,"-a")){
+            RenameAssociatedFiles = TRUE;
+            #ifndef _WIN32
+                ErrFatal("Error: -a only supported in Windows version");
+            #endif 
         }else if (!strcmp(arg,"-ft")){
             Exif2FileTime = TRUE;
             DoReadAction = TRUE;
@@ -1140,12 +1384,12 @@ int main (int argc, char **argv)
             int hours, minutes, seconds, n;
             minutes = seconds = 0;
             if (arg[3] != '-' && arg[3] != '+'){
-                ErrFatal("Error: -ta must be followed by +/- and a time\n");
+                ErrFatal("Error: -ta must be followed by +/- and a time");
             }
             n = sscanf(arg+4, "%d:%d:%d", &hours, &minutes, &seconds);
 
             if (n < 1){
-                ErrFatal("Error: -ta must be immediately followed by time\n");
+                ErrFatal("Error: -ta must be immediately followed by time");
             }
             if (ExifTimeAdjust) ErrFatal("Can only use one of -da or -ta options at once");
             ExifTimeAdjust = hours*3600 + minutes*60 + seconds;
@@ -1176,7 +1420,7 @@ int main (int argc, char **argv)
             
             if (!Exif2tm(&tm, arg+3)){
                 ErrFatal("-ts option must be followed by time in format yyyy:mmm:dd-hh:mm:ss\n"
-                        "Example: jhead -ts2001:01:01-12:00:00 foo.jpg\n");
+                        "Example: jhead -ts2001:01:01-12:00:00 foo.jpg");
             }
 
             ExifTimeSet  = mktime(&tm);
@@ -1184,47 +1428,21 @@ int main (int argc, char **argv)
             if ((int)ExifTimeSet == -1) ErrFatal("Time specified is out of range");
             DoModify = TRUE;
 
-        }else if (!strcmp(arg,"-cmd")){
-            if (argn+1 >= argc) Usage(); // No extra argument.
-            ApplyCommand = argv[++argn];
-            DoModify = TRUE;
+    // File matching and selection
         }else if (!strcmp(arg,"-model")){
             if (argn+1 >= argc) Usage(); // No extra argument.
             FilterModel = argv[++argn];
         }else if (!strcmp(arg,"-exonly")){
             ExifOnly = 1;
-        }else if (!strcmp(arg,"-c")){
-            ShowConcise = TRUE;
-        }else if (!strcmp(arg,"-h")){
-            Usage();
-        }else if (!strcmp(arg,"-dc")){
-            DeleteComments = TRUE;
+        }else if (!strcmp(arg,"-orp")){
+            PortraitOnly = 1;
+        }else if (!strcmp(arg,"-orl")){
+            PortraitOnly = -1;
+        }else if (!strcmp(arg,"-cmd")){
+            if (argn+1 >= argc) Usage(); // No extra argument.
+            ApplyCommand = argv[++argn];
             DoModify = TRUE;
-        }else if (!strcmp(arg,"-de")){
-            DeleteExif = TRUE;
-            DoModify = TRUE;
-        }else if (!strcmp(arg, "-du")){
-            DeleteUnknown = TRUE;
-            DoModify = TRUE;
-        }else if (!strcmp(arg, "-purejpg")){
-            DeleteExif = TRUE;
-            DeleteComments = TRUE;
-            DeleteUnknown = TRUE;
-            DoModify = TRUE;
-        }else if (!strcmp(arg,"-ce")){
-            EditComment = TRUE;
-            DoModify = TRUE;
-        }else if (!strcmp(arg,"-cs")){
-            CommentSavefileName = argv[++argn];
-            DoModify = TRUE;
-        }else if (!strcmp(arg,"-ci")){
-            CommentInsertfileName = argv[++argn];
-            DoModify = TRUE;
-        }else if (!strcmp(arg,"-cl")){
-            CommentInsertLiteral = argv[++argn];
-            DoModify = TRUE;
-        }else if (!strcmp(arg,"-se")){
-            SupressNonFatalErrors = TRUE;
+
 #ifdef MATTHIAS
         }else if (!strcmp(arg,"-ca")){
             // Its a literal comment.  Add.
@@ -1246,7 +1464,7 @@ int main (int argc, char **argv)
             exit(-1);
         }
         if (argn >= argc){
-            // Used an extr argument - becuase the last argument 
+            // Used an extra argument - becuase the last argument 
             // used up an extr argument.
             ErrFatal("Extra argument required");
         }
@@ -1255,12 +1473,18 @@ int main (int argc, char **argv)
         ErrFatal("No files to process.  Use -h for help");
     }
 
-    if (ThumbnailName != NULL && strcmp(ThumbnailName, "&i") == 0){
+    if (ThumbSaveName != NULL && strcmp(ThumbSaveName, "&i") == 0){
         printf("Error: By specifying \"&i\" for the thumbail name, your original file\n"
                "       will be overwitten.  If this is what you really want,\n"
                "       specify  -st \"./&i\"  to override this check\n");
         exit(0);
-        
+    }
+
+    if (RegenThumbnail){
+        if (ThumbSaveName || ThumbInsertName){
+            printf("Error: Cannot regen and save or insert thumbnail in same run\n");
+            exit(0);
+        }
     }
 
     if (EditComment){
@@ -1274,7 +1498,7 @@ int main (int argc, char **argv)
     if (ExifXferScrFile){
         if (FilterModel || ApplyCommand){
             ErrFatal("Error: Filter by model and/or applying command to files\n"
-            "   invalid while transfering Exif headers\n");
+            "   invalid while transfering Exif headers");
         }
     }
 
@@ -1290,14 +1514,12 @@ int main (int argc, char **argv)
                     if (argv[argn][a] == '/') argv[argn][a] = '\\';
                 }
             }
-
             // Use my globbing module to do fancier wildcard expansion with recursive
             // subdirectories under Windows.
-
             MyGlob(argv[argn], ProcessFile);
         #else
             // Under linux, don't do any extra fancy globbing - shell globbing is 
-            // pretty fancy as it is.
+            // pretty fancy as it is - although not as good as myglob.c
             ProcessFile(argv[argn]);
         #endif
 
@@ -1313,3 +1535,4 @@ int main (int argc, char **argv)
     }
 }
 
+
